@@ -9,32 +9,31 @@
 #include <filesystem>
 
 bool waylandEnums = false;
+bool clientCode   = false;
+bool noInterfaces = false;
 
 struct SRequestArgument {
     std::string wlType;
     std::string interface;
     std::string enumName;
     std::string name;
+    bool        newType   = false;
     bool        allowNull = false;
 };
 
-struct SRequest {
-    std::vector<SRequestArgument> args;
-    std::string                   name;
-    std::string                   since;
-};
-
-struct SEvent {
+struct SWaylandFunction {
     std::vector<SRequestArgument> args;
     std::string                   name;
     std::string                   since;
+    std::string                   newIdType  = ""; // client only
+    bool                          destructor = false;
 };
 
 struct SInterface {
-    std::vector<SRequest> requests;
-    std::vector<SEvent>   events;
-    std::string           name;
-    int                   version = 1;
+    std::vector<SWaylandFunction> requests;
+    std::vector<SWaylandFunction> events;
+    std::string                   name;
+    int                           version = 1;
 };
 
 struct SEnum {
@@ -108,17 +107,22 @@ std::string WPTypeToCType(const SRequestArgument& arg, bool event /* events pass
             return "uint32_t";
 
         // enum
-        if (!arg.enumName.empty())
+        if (!arg.enumName.empty()) {
             for (auto& e : XMLDATA.enums) {
                 if (e.nameOriginal == arg.enumName)
                     return e.name;
             }
+            return "uint32_t";
+        }
+
+        if (!event && clientCode && arg.wlType == "new_id")
+            return "wl_proxy*";
 
         // iface
         if (!arg.interface.empty() && event) {
             for (auto& i : XMLDATA.ifaces) {
                 if (i.name == arg.interface)
-                    return camelize("C_" + arg.interface + "*");
+                    return camelize((clientCode ? "CC_" : "C_") + arg.interface + "*");
             }
             return "wl_resource*";
         }
@@ -129,7 +133,7 @@ std::string WPTypeToCType(const SRequestArgument& arg, bool event /* events pass
         if (!arg.interface.empty() && event && !ignoreTypes) {
             for (auto& i : XMLDATA.ifaces) {
                 if (i.name == arg.interface)
-                    return camelize("C_" + arg.interface + "*");
+                    return camelize((clientCode ? "CC_" : "C_") + arg.interface + "*");
             }
         }
         return "wl_resource*";
@@ -187,12 +191,17 @@ void parseXML(pugi::xml_document& doc) {
         }
 
         for (auto& rq : iface.children("request")) {
-            SRequest srq;
-            srq.name  = rq.attribute("name").as_string();
-            srq.since = rq.attribute("since").as_string();
+            SWaylandFunction srq;
+            srq.name       = rq.attribute("name").as_string();
+            srq.since      = rq.attribute("since").as_string();
+            srq.destructor = rq.attribute("type").as_string() == std::string{"destructor"};
 
             for (auto& arg : rq.children("arg")) {
                 SRequestArgument sargm;
+                if (arg.attribute("type").as_string() == std::string{"new_id"} && clientCode)
+                    srq.newIdType = arg.attribute("interface").as_string();
+
+                sargm.newType   = arg.attribute("type").as_string() == std::string{"new_id"} && clientCode;
                 sargm.name      = sanitize(arg.attribute("name").as_string());
                 sargm.wlType    = arg.attribute("type").as_string();
                 sargm.interface = arg.attribute("interface").as_string();
@@ -206,9 +215,10 @@ void parseXML(pugi::xml_document& doc) {
         }
 
         for (auto& ev : iface.children("event")) {
-            SEvent sev;
-            sev.name  = ev.attribute("name").as_string();
-            sev.since = ev.attribute("since").as_string();
+            SWaylandFunction sev;
+            sev.name       = ev.attribute("name").as_string();
+            sev.since      = ev.attribute("since").as_string();
+            sev.destructor = ev.attribute("type").as_string() == std::string{"destructor"};
 
             for (auto& arg : ev.children("arg")) {
                 SRequestArgument sargm;
@@ -231,19 +241,20 @@ void parseXML(pugi::xml_document& doc) {
 void parseHeader() {
 
     // add some boilerplate
-    HEADER += R"#(#pragma once
+    HEADER += std::format(R"#(#pragma once
 
 #include <functional>
 #include <cstdint>
 #include <string>
-#include <wayland-server.h>
+{}
 
 #define F std::function
 
-struct wl_client;
-struct wl_resource;
+{}
 
-)#";
+)#",
+                          (clientCode ? "#include <wayland-client.h>" : "#include <wayland-server.h>"),
+                          (clientCode ? "struct wl_proxy;\ntypedef wl_proxy wl_resource;" : "struct wl_client;\nstruct wl_resource;"));
 
     // parse all enums
     if (!waylandEnums) {
@@ -258,13 +269,13 @@ struct wl_resource;
 
     // fw declare all classes
     for (auto& iface : XMLDATA.ifaces) {
-        const auto IFACE_CLASS_NAME_CAMEL = camelize("C_" + iface.name);
+        const auto IFACE_CLASS_NAME_CAMEL = camelize((clientCode ? "CC_" : "C_") + iface.name);
         HEADER += std::format("\nclass {};", IFACE_CLASS_NAME_CAMEL);
 
         for (auto& rq : iface.requests) {
             for (auto& arg : rq.args) {
                 if (!arg.interface.empty()) {
-                    HEADER += std::format("\nclass {};", camelize("C_" + arg.interface));
+                    HEADER += std::format("\nclass {};", camelize((clientCode ? "CC_" : "C_") + arg.interface));
                 }
             }
         }
@@ -272,7 +283,7 @@ struct wl_resource;
         for (auto& rq : iface.events) {
             for (auto& arg : rq.args) {
                 if (!arg.interface.empty()) {
-                    HEADER += std::format("\nclass {};", camelize("C_" + arg.interface));
+                    HEADER += std::format("\nclass {};", camelize((clientCode ? "CC_" : "C_") + arg.interface));
                 }
             }
         }
@@ -291,15 +302,32 @@ struct wl_resource;
 
     for (auto& iface : XMLDATA.ifaces) {
         const auto IFACE_NAME_CAMEL       = camelize(iface.name);
-        const auto IFACE_CLASS_NAME_CAMEL = camelize("C_" + iface.name);
+        const auto IFACE_CLASS_NAME_CAMEL = camelize((clientCode ? "CC_" : "C_") + iface.name);
+
+        if (!clientCode) {
+            HEADER += std::format(R"#(
+struct {}DestroyWrapper {{
+    wl_listener listener;
+    {}* parent = nullptr;
+}};
+            )#",
+                                  IFACE_CLASS_NAME_CAMEL, IFACE_CLASS_NAME_CAMEL);
+        }
 
         // begin the class
-        HEADER += std::format(R"#(
+        HEADER +=
+            std::format(R"#(
+
 class {} {{
   public:
-    {}(wl_client* client, uint32_t version, uint32_t id);
+    {}({});
     ~{}();
 
+)#",
+                        IFACE_CLASS_NAME_CAMEL, IFACE_CLASS_NAME_CAMEL, (clientCode ? "wl_resource*" : "wl_client* client, uint32_t version, uint32_t id"), IFACE_CLASS_NAME_CAMEL);
+
+        if (!clientCode) {
+            HEADER += std::format(R"#(
     // set a listener for when this resource is _being_ destroyed
     void setOnDestroy(F<void({}*)> handler) {{
         onDestroy = handler;
@@ -314,7 +342,7 @@ class {} {{
     void* data() {{
         return pData;
     }}
-    
+
     // get the raw wl_resource ptr
     wl_resource* resource() {{
         return pResource;
@@ -339,17 +367,41 @@ class {} {{
     int version() {{
         return wl_resource_get_version(pResource);
     }}
+            )#",
+                                  IFACE_CLASS_NAME_CAMEL);
+        } else {
+            HEADER += R"#(
+    // set the data for this resource
+    void setData(void* data) {{
+        pData = data;
+    }}
 
-)#",
-                              IFACE_CLASS_NAME_CAMEL, IFACE_CLASS_NAME_CAMEL, IFACE_CLASS_NAME_CAMEL, IFACE_CLASS_NAME_CAMEL);
+    // get the data for this resource
+    void* data() {{
+        return pData;
+    }}
+
+    // get the raw wl_resource (wl_proxy) ptr
+    wl_resource* resource() {{
+        return pResource;
+    }}
+
+    // get the resource version
+    int version() {{
+        return wl_proxy_get_version(pResource);
+    }}
+            )#";
+        }
 
         // add all setters for requests
         HEADER += "\n    // --------------- Requests --------------- //\n\n";
 
-        for (auto& rq : iface.requests) {
+        for (auto& rq : (clientCode ? iface.events : iface.requests)) {
 
             std::string args = ", ";
             for (auto& arg : rq.args) {
+                if (arg.newType)
+                    continue;
                 args += WPTypeToCType(arg, false) + ", ";
             }
 
@@ -363,9 +415,11 @@ class {} {{
 
         HEADER += "\n    // --------------- Events --------------- //\n\n";
 
-        for (auto& ev : iface.events) {
+        for (auto& ev : (!clientCode ? iface.events : iface.requests)) {
             std::string args = "";
             for (auto& arg : ev.args) {
+                if (arg.newType)
+                    continue;
                 args += WPTypeToCType(arg, true) + ", ";
             }
 
@@ -374,22 +428,26 @@ class {} {{
                 args.pop_back();
             }
 
-            HEADER += std::format("    void {}({});\n", camelize("send_" + ev.name), args);
+            HEADER += std::format("    {} {}({});\n", ev.newIdType.empty() ? "void" : "wl_proxy*", camelize("send_" + ev.name), args);
         }
 
         // dangerous ones
-        for (auto& ev : iface.events) {
-            std::string args = "";
-            for (auto& arg : ev.args) {
-                args += WPTypeToCType(arg, true, true) + ", ";
-            }
+        if (!clientCode) {
+            for (auto& ev : (!clientCode ? iface.events : iface.requests)) {
+                std::string args = "";
+                for (auto& arg : ev.args) {
+                    if (arg.newType)
+                        continue;
+                    args += WPTypeToCType(arg, true, true) + ", ";
+                }
 
-            if (!args.empty()) {
-                args.pop_back();
-                args.pop_back();
-            }
+                if (!args.empty()) {
+                    args.pop_back();
+                    args.pop_back();
+                }
 
-            HEADER += std::format("    void {}({});\n", camelize("send_" + ev.name + "_raw"), args);
+                HEADER += std::format("    void {}({});\n", camelize("send_" + ev.name + "_raw"), args);
+            }
         }
 
         // end events
@@ -400,10 +458,12 @@ class {} {{
         // start requests storage
         HEADER += "    struct {\n";
 
-        for (auto& rq : iface.requests) {
+        for (auto& rq : (clientCode ? iface.events : iface.requests)) {
 
             std::string args = ", ";
             for (auto& arg : rq.args) {
+                if (arg.newType)
+                    continue;
                 args += WPTypeToCType(arg, false) + ", ";
             }
 
@@ -419,17 +479,26 @@ class {} {{
         HEADER += "    } requests;\n";
 
         // constant resource stuff
-        HEADER += std::format(R"#(
+        if (!clientCode) {
+            HEADER += std::format(R"#(
     void onDestroyCalled();
 
     F<void({}*)> onDestroy;
 
     wl_resource* pResource = nullptr;
 
-    wl_listener resourceDestroyListener;
+    {}DestroyWrapper resourceDestroyListener;
 
     void* pData = nullptr;)#",
-                              IFACE_CLASS_NAME_CAMEL);
+                                  IFACE_CLASS_NAME_CAMEL, IFACE_CLASS_NAME_CAMEL);
+        } else {
+            HEADER += R"#(
+    wl_resource* pResource = nullptr;
+
+    bool destroyed = false;
+
+    void* pData = nullptr;)#";
+        }
 
         HEADER += "\n};\n\n";
     }
@@ -510,14 +579,16 @@ static const wl_interface* dummyTypes[] = { nullptr };
         const auto IFACE_WL_NAME          = iface.name + "_interface";
         const auto IFACE_NAME             = iface.name;
         const auto IFACE_NAME_CAMEL       = camelize(iface.name);
-        const auto IFACE_CLASS_NAME_CAMEL = camelize("C_" + iface.name);
+        const auto IFACE_CLASS_NAME_CAMEL = camelize((clientCode ? "CC_" : "C_") + iface.name);
 
         // create handlers
-        for (auto& rq : iface.requests) {
+        for (auto& rq : (clientCode ? iface.events : iface.requests)) {
             const auto  REQUEST_NAME = camelize(std::string{"_"} + "C_" + IFACE_NAME + "_" + rq.name);
 
             std::string argsC = ", ";
             for (auto& arg : rq.args) {
+                if (arg.newType)
+                    continue;
                 argsC += WPTypeToCType(arg, false) + " " + arg.name + ", ";
             }
 
@@ -534,24 +605,38 @@ static const wl_interface* dummyTypes[] = { nullptr };
                 argsN.pop_back();
             }
 
-            SOURCE += std::format(R"#(
+            if (!clientCode) {
+                SOURCE += std::format(R"#(
 static void {}(wl_client* client, wl_resource* resource{}) {{
     const auto PO = ({}*)wl_resource_get_user_data(resource);
     if (PO && PO->requests.{})
         PO->requests.{}(PO{});
 }}
 )#",
-                                  REQUEST_NAME, argsC, IFACE_CLASS_NAME_CAMEL, camelize(rq.name), camelize(rq.name), argsN);
+                                      REQUEST_NAME, argsC, IFACE_CLASS_NAME_CAMEL, camelize(rq.name), camelize(rq.name), argsN);
+            } else {
+                SOURCE += std::format(R"#(
+static void {}(void* data, void* resource{}) {{
+    const auto PO = ({}*)data;
+    if (PO && PO->requests.{})
+        PO->requests.{}(PO{});
+}}
+)#",
+                                      REQUEST_NAME, argsC, IFACE_CLASS_NAME_CAMEL, camelize(rq.name), camelize(rq.name), argsN);
+            }
         }
 
         // destroy handler
-        SOURCE += std::format(R"#(
+        if (!clientCode) {
+            SOURCE += std::format(R"#(
 static void _{}__DestroyListener(wl_listener* l, void* d) {{
-    {}* pResource = wl_container_of(l, pResource, resourceDestroyListener);
+    {}DestroyWrapper *wrap = wl_container_of(l, wrap, listener);
+    {}* pResource = wrap->parent;
     pResource->onDestroyCalled();
 }}
 )#",
-                              IFACE_CLASS_NAME_CAMEL, IFACE_CLASS_NAME_CAMEL);
+                                  IFACE_CLASS_NAME_CAMEL, IFACE_CLASS_NAME_CAMEL, IFACE_CLASS_NAME_CAMEL);
+        }
 
         // create vtable
 
@@ -562,7 +647,7 @@ static const void* {}[] = {{
 )#",
                               IFACE_VTABLE_NAME);
 
-        for (auto& rq : iface.requests) {
+        for (auto& rq : (clientCode ? iface.events : iface.requests)) {
             const auto REQUEST_NAME = camelize(std::string{"_"} + "C_" + IFACE_NAME + "_" + rq.name);
             SOURCE += std::format("    (void*){},\n", REQUEST_NAME);
         }
@@ -572,11 +657,13 @@ static const void* {}[] = {{
         // create events
 
         int evid = 0;
-        for (auto& ev : iface.events) {
+        for (auto& ev : (!clientCode ? iface.events : iface.requests)) {
             const auto  EVENT_NAME = camelize("send_" + ev.name);
 
             std::string argsC = "";
             for (auto& arg : ev.args) {
+                if (arg.newType)
+                    continue;
                 argsC += WPTypeToCType(arg, true) + " " + arg.name + ", ";
             }
 
@@ -587,60 +674,85 @@ static const void* {}[] = {{
 
             std::string argsN = ", ";
             for (auto& arg : ev.args) {
+                if (arg.newType)
+                    continue;
                 if (!WPTypeToCType(arg, true).starts_with("C"))
                     argsN += arg.name + ", ";
                 else
-                    argsN += arg.name + "->pResource, ";
+                    argsN += (arg.name + " ? " + arg.name + "->pResource : nullptr, ");
             }
 
             argsN.pop_back();
             argsN.pop_back();
 
-            SOURCE += std::format(R"#(
+            if (!clientCode) {
+                SOURCE += std::format(R"#(
 void {}::{}({}) {{
     if (!pResource)
         return;
     wl_resource_post_event(pResource, {}{});
 }}
 )#",
-                                  IFACE_CLASS_NAME_CAMEL, EVENT_NAME, argsC, evid, argsN);
+                                      IFACE_CLASS_NAME_CAMEL, EVENT_NAME, argsC, evid, argsN);
+            } else {
+                std::string retType    = ev.newIdType.empty() ? "void" : "wl_proxy";
+                std::string ptrRetType = ev.newIdType.empty() ? "void" : "wl_proxy*";
+                std::string flags      = ev.destructor ? "1" : "0";
+                SOURCE += std::format(R"#(
+{} {}::{}({}) {{
+    if (!pResource)
+        return{};{}
+
+    auto proxy = wl_proxy_marshal_flags((wl_proxy*)pResource, {}, {}, wl_proxy_get_version((wl_proxy*)pResource), {}{}{});{}
+}}
+)#",
+                                      ptrRetType, IFACE_CLASS_NAME_CAMEL, EVENT_NAME, argsC, (ev.newIdType.empty() ? "" : " nullptr"),
+                                      (ev.destructor ? "\n    destroyed = true;" : ""), evid, (ev.newIdType.empty() ? "nullptr" : "&" + ev.newIdType + "_interface"), flags,
+                                      (!ev.newIdType.empty() ? ", nullptr" : ""), argsN, (ev.newIdType.empty() ? "\n    proxy;" : "\n\n    return proxy;"));
+            }
 
             evid++;
         }
 
         // dangerous
-        evid = 0;
-        for (auto& ev : iface.events) {
-            const auto  EVENT_NAME = camelize("send_" + ev.name + "_raw");
+        if (!clientCode) {
+            evid = 0;
+            for (auto& ev : iface.events) {
+                const auto  EVENT_NAME = camelize("send_" + ev.name + "_raw");
 
-            std::string argsC = "";
-            for (auto& arg : ev.args) {
-                argsC += WPTypeToCType(arg, true, true) + " " + arg.name + ", ";
-            }
+                std::string argsC = "";
+                for (auto& arg : ev.args) {
+                    if (arg.newType)
+                        continue;
+                    argsC += WPTypeToCType(arg, true, true) + " " + arg.name + ", ";
+                }
 
-            if (!argsC.empty()) {
-                argsC.pop_back();
-                argsC.pop_back();
-            }
+                if (!argsC.empty()) {
+                    argsC.pop_back();
+                    argsC.pop_back();
+                }
 
-            std::string argsN = ", ";
-            for (auto& arg : ev.args) {
-                argsN += arg.name + ", ";
-            }
+                std::string argsN = ", ";
+                for (auto& arg : ev.args) {
+                    if (arg.newType)
+                        continue;
+                    argsN += arg.name + ", ";
+                }
 
-            argsN.pop_back();
-            argsN.pop_back();
+                argsN.pop_back();
+                argsN.pop_back();
 
-            SOURCE += std::format(R"#(
+                SOURCE += std::format(R"#(
 void {}::{}({}) {{
     if (!pResource)
         return;
     wl_resource_post_event(pResource, {}{});
 }}
 )#",
-                                  IFACE_CLASS_NAME_CAMEL, EVENT_NAME, argsC, evid, argsN);
+                                      IFACE_CLASS_NAME_CAMEL, EVENT_NAME, argsC, evid, argsN);
 
-            evid++;
+                evid++;
+            }
         }
 
         // wayland interfaces and stuff
@@ -687,49 +799,52 @@ void {}::{}({}) {{
         const auto MESSAGE_NAME_EVENTS   = camelize(std::string{"_"} + "C_" + IFACE_NAME + "_events");
 
         // message
-        if (iface.requests.size() > 0) {
-            SOURCE += std::format(R"#(
+        if (!noInterfaces) {
+            if (iface.requests.size() > 0) {
+                SOURCE += std::format(R"#(
 static const wl_message {}[] = {{
 )#",
-                                  MESSAGE_NAME_REQUESTS);
-            for (auto& rq : iface.requests) {
-                // create type table
-                const auto TYPE_TABLE_NAME = camelize(std::string{"_"} + "C_" + IFACE_NAME + "_" + rq.name + "_types");
+                                      MESSAGE_NAME_REQUESTS);
+                for (auto& rq : iface.requests) {
+                    // create type table
+                    const auto TYPE_TABLE_NAME = camelize(std::string{"_"} + "C_" + IFACE_NAME + "_" + rq.name + "_types");
 
-                SOURCE += std::format("    {{ \"{}\", \"{}\", {}}},\n", rq.name, argsToShort(rq.args, rq.since), rq.args.empty() ? "dummyTypes + 0" : TYPE_TABLE_NAME + " + 0");
+                    SOURCE += std::format("    {{ \"{}\", \"{}\", {}}},\n", rq.name, argsToShort(rq.args, rq.since), rq.args.empty() ? "dummyTypes + 0" : TYPE_TABLE_NAME + " + 0");
+                }
+
+                SOURCE += "};\n";
             }
 
-            SOURCE += "};\n";
-        }
-
-        if (iface.events.size() > 0) {
-            SOURCE += std::format(R"#(
+            if (iface.events.size() > 0) {
+                SOURCE += std::format(R"#(
 static const wl_message {}[] = {{
 )#",
-                                  MESSAGE_NAME_EVENTS);
-            for (auto& ev : iface.events) {
-                // create type table
-                const auto TYPE_TABLE_NAME = camelize(std::string{"_"} + "C_" + IFACE_NAME + "_" + ev.name + "_types");
+                                      MESSAGE_NAME_EVENTS);
+                for (auto& ev : iface.events) {
+                    // create type table
+                    const auto TYPE_TABLE_NAME = camelize(std::string{"_"} + "C_" + IFACE_NAME + "_" + ev.name + "_types");
 
-                SOURCE += std::format("    {{ \"{}\", \"{}\", {}}},\n", ev.name, argsToShort(ev.args, ev.since), ev.args.empty() ? "dummyTypes + 0" : TYPE_TABLE_NAME + " + 0");
+                    SOURCE += std::format("    {{ \"{}\", \"{}\", {}}},\n", ev.name, argsToShort(ev.args, ev.since), ev.args.empty() ? "dummyTypes + 0" : TYPE_TABLE_NAME + " + 0");
+                }
+
+                SOURCE += "};\n";
             }
 
-            SOURCE += "};\n";
-        }
-
-        // iface
-        SOURCE += std::format(R"#(
+            // iface
+            SOURCE += std::format(R"#(
 const wl_interface {} = {{
     "{}", {},
     {}, {},
     {}, {},
 }};
 )#",
-                              IFACE_WL_NAME, iface.name, iface.version, iface.requests.size(), (iface.requests.size() > 0 ? MESSAGE_NAME_REQUESTS : "nullptr"), iface.events.size(),
-                              (iface.events.size() > 0 ? MESSAGE_NAME_EVENTS : "nullptr"));
+                                  IFACE_WL_NAME, iface.name, iface.version, iface.requests.size(), (iface.requests.size() > 0 ? MESSAGE_NAME_REQUESTS : "nullptr"),
+                                  iface.events.size(), (iface.events.size() > 0 ? MESSAGE_NAME_EVENTS : "nullptr"));
+        }
 
         // protocol body
-        SOURCE += std::format(R"#(
+        if (!clientCode) {
+            SOURCE += std::format(R"#(
 {}::{}(wl_client* client, uint32_t version, uint32_t id) {{
     pResource = wl_resource_create(client, &{}, version, id);
 
@@ -737,16 +852,17 @@ const wl_interface {} = {{
         return;
 
     wl_resource_set_user_data(pResource, this);
-    wl_list_init(&resourceDestroyListener.link);
-    resourceDestroyListener.notify = _{}__DestroyListener;
-    wl_resource_add_destroy_listener(pResource, &resourceDestroyListener);
+    wl_list_init(&resourceDestroyListener.listener.link);
+    resourceDestroyListener.listener.notify = _{}__DestroyListener;
+    resourceDestroyListener.parent = this;
+    wl_resource_add_destroy_listener(pResource, &resourceDestroyListener.listener);
 
     wl_resource_set_implementation(pResource, {}, this, nullptr);
 }}
 
 {}::~{}() {{
-    wl_list_remove(&resourceDestroyListener.link);
-    wl_list_init(&resourceDestroyListener.link);
+    wl_list_remove(&resourceDestroyListener.listener.link);
+    wl_list_init(&resourceDestroyListener.listener.link);
 
     // if we still own the wayland resource,
     // it means we need to destroy it.
@@ -758,8 +874,8 @@ const wl_interface {} = {{
 
 void {}::onDestroyCalled() {{
     wl_resource_set_user_data(pResource, nullptr);
-    wl_list_remove(&resourceDestroyListener.link);
-    wl_list_init(&resourceDestroyListener.link);
+    wl_list_remove(&resourceDestroyListener.listener.link);
+    wl_list_init(&resourceDestroyListener.listener.link);
 
     // set the resource to nullptr,
     // as it will be freed. If the consumer does not destroy this resource
@@ -770,10 +886,41 @@ void {}::onDestroyCalled() {{
         onDestroy(this);
 }}
 )#",
-                              IFACE_CLASS_NAME_CAMEL, IFACE_CLASS_NAME_CAMEL, IFACE_NAME + "_interface", IFACE_CLASS_NAME_CAMEL, IFACE_VTABLE_NAME, IFACE_CLASS_NAME_CAMEL,
-                              IFACE_CLASS_NAME_CAMEL, IFACE_CLASS_NAME_CAMEL);
+                                  IFACE_CLASS_NAME_CAMEL, IFACE_CLASS_NAME_CAMEL, IFACE_NAME + "_interface", IFACE_CLASS_NAME_CAMEL, IFACE_VTABLE_NAME, IFACE_CLASS_NAME_CAMEL,
+                                  IFACE_CLASS_NAME_CAMEL, IFACE_CLASS_NAME_CAMEL);
+        } else {
+            std::string DTOR_FUNC = "";
 
-        for (auto& rq : iface.requests) {
+            for (auto& rq : iface.requests) {
+                if (!rq.destructor)
+                    continue;
+
+                DTOR_FUNC = camelize("send_" + rq.name) + "()";
+                break;
+            }
+
+            if (DTOR_FUNC.empty())
+                DTOR_FUNC = "wl_proxy_destroy(pResource)";
+
+            SOURCE += std::format(R"#(
+{}::{}(wl_resource* resource) {{
+    pResource = resource;
+
+    if (!pResource)
+        return;
+
+    wl_proxy_add_listener(pResource, (void (**)(void))&{}, this);
+}}
+
+{}::~{}() {{
+    if (!destroyed)
+        {};
+}}
+)#",
+                                  IFACE_CLASS_NAME_CAMEL, IFACE_CLASS_NAME_CAMEL, IFACE_VTABLE_NAME, IFACE_CLASS_NAME_CAMEL, IFACE_CLASS_NAME_CAMEL, DTOR_FUNC);
+        }
+
+        for (auto& rq : (clientCode ? iface.events : iface.requests)) {
             std::string args = ", ";
             for (auto& arg : rq.args) {
                 args += WPTypeToCType(arg, false) + ", ";
@@ -808,6 +955,16 @@ int main(int argc, char** argv, char** envp) {
             return 0;
         }
 
+        if (curarg == "-c" || curarg == "--client") {
+            clientCode = true;
+            continue;
+        }
+
+        if (curarg == "--no-interfaces") {
+            noInterfaces = true;
+            continue;
+        }
+
         if (curarg == "--wayland-enums") {
             waylandEnums = true;
             continue;
